%{
/* syntax analyzer for asm3v ver.1.0 ['92/Aug/5-11] */
/* 			     ver.1.1 ['92/Aug/22]   */
/* 			     ver.1.2 ['92/Aug/24]   */
/* 			     ver.1.3 ['92/Sep/1-2]  */
/* 			     ver.1.4 ['92/Sep/3]   */
/* 			     ver.1.5 ['92/Sep/9]   */
/* 			     ver.1.6 ['93/Feb/5]   */
/* 			     ver.1.6R['93/Oct/26]  */

/* syntax analyzer for asm4v ver.1.0 ['94/Jun/10] */
/* 			     ver.1.1 ['94/Dec/19] */
/* 			     ver.1.3 ['96/May/22] */

#include <stdio.h>
#include <malloc.h>
#include "defines.h"
#include "ipconst.h"
#include "opcode.h"
#include "regdef.h"
#include "specialreg.h"

#define	ZR	0
#define	PC	N_PC
#define	FR	N_FR

/* chainet offset (for identify CN) */
#define CNOFFSET 100

/* DON'T forget check HALT value in defines.h(ps3v) !!! */
#define HALT	(1<<8)

/* bit mask pattern */
static int bitmask[]={  0,
			0x1,       0x3,       0x7,       0xf,
			0x1f,      0x3f,      0x7f,      0xff,
			0x1ff,     0x3ff,     0x7ff,     0xfff,
			0x1fff,    0x3fff,    0x7fff,    0xffff,
			0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
			0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
			0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
			0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

/* enable uninitialized label used error flag */
int	uninitlabelerr = FALSE;

/* instruction struct */
typedef struct {
	int	optype;
	int	subtype1;
	int	subtype2;
	int	subtype3;
	int	src1;
	int	src2;
	int	dstn;
	int	vmr;
} insttype;

insttype inst;

/* composed instruction word */
int	instword;

/* PC increment value */
int	incpc;

/* DW pseudo instruction counter */
int	dwcount;

/* calculate relative address */
int	reladdr();

/* check bit width */
int	checkwidth();

/* make chained vector instruction operands */
int	makecvinst();

/* error print routine */
void	yyerror();

/* error flag */
int	errorflag = FALSE;

%}
%union {
	int	intvalue;
	float	floatvalue;
	char	*charptr;
}
%left OROP
%left XOROP
%left ANDOP
%left RSFTOP LSFTOP
%left ADDOP SUBOP
%left MULOP DIVOP MODOP
%right UMINUS UPLUS NOTOP
%token LP RP
%token LDI LD LDB LDBI LDBS LDPCB LDPCBI
%token ST STB STBI STBS STPCB STPCBI
%token ADD ADDI SUB SUBI MUL MULI
%token DIV DIVI MOD MODI RSUB RSUBI
%token UMUL UMULI UDIV UDIVI AND ANDI
%token OR ORI XOR XORI NOT SRL SRLI
%token SLL SLLI SRA SRAI SLA SLAI
%token FADD FSUB FMUL FDIV INT FLT
%token JMP JC JB JBC
%token JMPL JCL JBL JBCL
%token CJP CJM CJZ CJNZ CJPI CJMI CJZI CJNZI
%token VFADD VFSUB VFMUL VFDIV
%token VADD VSUB VMUL VDIV
%token VAND VOR VXOR VNOT
%token CVFADD CVFSUB CVFMUL CVFDIV
%token CVADD CVSUB CVMUL CVDIV
%token CVAND CVOR CVXOR CVNOT
%token VBCAST VSRL VSLL
%token VLIR VLII VSIR VSII LVLD LVST LVBLD LVBST
%token VMRSETI VMRSETF VMRAND VMROR VMRXOR VMRNOT
%token VINT VFLT MOVVS MOVSV
%token P M Z NP NM NZ
%token ORG EQU DW DS
%token NOP HLT
%token MOV CMP CMPI RCMP RCMPI INC DEC CLR
%token CODE DATA
%token COLON COMMA
%token <intvalue> SR VR CN VMR INTVAL SR_L SR_G SR_F SR_B
%token <floatvalue> FLOATVAL
%token <charptr> LABEL STRING NL
%type <intvalue> intexp vrcn sr128 pseudoinst sreg
%start program
%%
program
	: linelist
	;

linelist
	: /* empty */
	| linelist line
	;

line
	: NL
		{
		if (pass == 2) {
		    listoutcmnt($1);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$1);
		    errorflag = FALSE;
		}
		}
	| LABEL COLON NL
		{
		if (pass == 1) {
		    if (asmmode == CODE_MODE) {
			if ((currentpc % NIP) != 0) {
			    yyerror("Code address label alignment error");
			}
		    }
		    addlabelandinit($1,currentpc);
		} else { /* pass 2 */
		    listoutwithpc(currentpc,$3);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$3);
		    errorflag = FALSE;
		}
		}
	| LABEL COLON instruction NL
		{
		if (pass == 1) {
		    if (asmmode == CODE_MODE) {
			if ((currentpc % NIP) != 0) {
			    yyerror("Code address label alignment error");
			}
		    }
		    addlabelandinit($1,currentpc);
		} else { /* pass 2 */
		    listoutwithinst(currentpc,instword,$4);
		    objectout(instword);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$4);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| instruction NL
		{
		if (pass == 2) {
		    listoutwithinst(currentpc,instword,$2);
		    objectout(instword);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$2);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| LABEL COLON pseudoinst NL
		{
		int	i;
		if (pass == 1) {
		    if (asmmode == CODE_MODE) {
			if ((currentpc % NIP) != 0) {
			    yyerror("Code address label alignment error");
			}
		    }
		    addlabelandinit($1,currentpc);
		} else { /* pass 2 */
		    switch ($3) {
			case ORG:
			    listoutwithpc(currentpc,$4);
			    objectaddrout(currentpc);
			    break;
			case DW:
			    listoutwithinst(currentpc,dwbuff[0].i,$4);
			    objectout(dwbuff[0].i);
			    for (i = 1 ; i < dwcount ; i++) {
				listoutwithinst(currentpc+i,
						dwbuff[i].i,
						"\n");
				objectout(dwbuff[i].i);
			    }
			    break;
			case DS:
			    listoutwithinst(currentpc,0,$4);
			    for (i = 0 ; i < incpc ; i++) {
				objectout(0);
			    }
			    break;
			case CODE:
			    listoutwithpc(currentpc,$4);
			    if (incpc != 0) {
				for (i = 0 ; i < incpc ; i++) {
				    objectout(0);
				}
			    }
			    break;
			case DATA:
			    listoutwithpc(currentpc,$4);
			    break;
			default:
			    listoutwithpc(currentpc,$4);
			    yyerror("irregal pseudo instruction");
			    break;
		    }
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$4);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| pseudoinst NL
		{
		int	i;
		if (pass == 2) {
		    switch ($1) {
			case ORG:
			    listoutwithpc(currentpc,$2);
			    objectaddrout(currentpc);
			    break;
			case DW:
			    listoutwithinst(currentpc,dwbuff[0].i,$2);
			    objectout(dwbuff[0].i);
			    for (i = 1 ; i < dwcount ; i++) {
				listoutwithinst(currentpc+i,
						dwbuff[i].i,
						"\n");
				objectout(dwbuff[i].i);
			    }
			    break;
			case DS:
			    listoutwithinst(currentpc,0,$2);
			    for (i = 0 ; i < incpc ; i++) {
				objectout(0);
			    }
			    break;
			case CODE:
			    listoutcmnt($2);
			    if (incpc != 0) {
				for (i = 0 ; i < incpc ; i++) {
				    objectout(0);
				}
			    }
			    break;
			case DATA:
			    listoutcmnt($2);
			    break;
			default:
			    listoutcmnt($2);
			    yyerror("irregal pseudo instruction");
			    break;
		    }
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$2);
		    errorflag = FALSE;
		}
		currentpc = currentpc + incpc;
		}
	| LABEL EQU {uninitlabelerr = TRUE;} intexp NL
		{
		if (pass == 1) {
		    addlabelandinit($1,$4);
		} else { /* pass 2 */
		    listoutequ(getlabelvalue($1),$5);
		}
		if (errorflag == TRUE) {
		    fprintf(stderr,$5);
		    errorflag = FALSE;
		}
		uninitlabelerr = uninitlabelerrdefault;
		}
	| error NL 
		{
		fprintf(stderr,$2);
		if (pass == 2) {
		    listoutcmnt($2);
		}
		errorflag = FALSE;
		yyerrok;
		}
	;

instruction
	: loadstore
	| RRRinst
	| IRRinst
	| RRinst
	| jump
	| compjump
	| vcalc
	| cvcalc
	| vsmixed
	| vls
	| lvls
	| vmrset
	| vmrop
	| vconv
	| vsmove
	| synthetic
	;

pseudoinst
	: ORG {uninitlabelerr = TRUE;} intexp
		{
		currentpc = $3;
		incpc = 0 ;
		uninitlabelerr = uninitlabelerrdefault;
		$$ = ORG;
		}
	| DW {dwcount = 0;} dwbody
		{
		if (pass == 1) {
			if (dwcount > maxdwc) maxdwc = dwcount;
			incpc = dwcount;
		} else { /* pass 2 */
			incpc = dwcount;
		}
		$$ = DW;
		}
	| DS {uninitlabelerr = TRUE;} intexp
		{
		incpc = $3 ;
		uninitlabelerr = uninitlabelerrdefault;
		$$ = DS;
		}
	| CODE
		{
		asmmode = CODE_MODE;
		if ((currentpc % NIP) != 0) {	/* instruction alignment */
			incpc = NIP - (currentpc % NIP);
		} else {
			incpc = 0 ;
		}
		$$ = CODE;
		}
	| DATA
		{
		asmmode = DATA_MODE;
		incpc = 0 ;
		$$ = DATA;
		}
	;

dwbody
	: dwterm
	| dwbody COMMA dwterm
	;

dwterm
	: intexp
		{
		if (pass == 1) {
			dwcount++;
		} else {
			dwbuff[dwcount++].i = $1;
		}
		}
	| STRING
		{
		if (pass == 1) {
			dwcount += strlen($1);
		} else {
			int i;
			for (i = 0 ; i < strlen($1) ; i++) {
				dwbuff[dwcount++].i = (int)$1[i];
			}
		}
		}
	| FLOATVAL
		{
		if (pass == 1) {
			dwcount++;
		} else {
			dwbuff[dwcount++].f = $1;
		}
		}
	;

sreg
	: SR
		{
		$$ = $1;
		}
	| SR_G 
		{
		$$ = $1;
		}
	| SR_L
		{
		int ip;

		ip = currentpc % NIP;
		$$ = GLOBAL+(LOCAL+OVERLAP*2)*ip+OVERLAP + $1;
		}
	| SR_B
		{
		int ip;

		ip = currentpc % NIP;
		$$ = GLOBAL+(LOCAL+OVERLAP*2)*ip + $1;
		}
	| SR_F
		{
		int ip;

		ip = currentpc % NIP;
		$$ = GLOBAL+(LOCAL+OVERLAP*2)*ip + OVERLAP + LOCAL + $1;
		}
	;

loadstore
	: loadimm intexp COMMA sreg
		{
		int	addrtmp;

		if (pass == 2) {
		    if (checkwidth($2,17)==TRUE) { /* normal LDI */ 
			instword =
			    (inst.optype << 29) |
			    (inst.subtype1 << 26) |
			    (($2 & 0x1ffff) << 8) |
			    $4 ;
		    } else {			  /* pseudo LDII */
			if (pldiicount>=MAXPLDII) {
			    yyerror("error : PseudoLDII buffer overflow");
			    $2 = 0;
			} else {
			    pldiivalue[pldiicount].i = $2;
			}
			/* PseudoLDI int = LOAD BASED(BASE=PSH)+DW */
			inst.optype = LOAD;
			inst.subtype1 = AD_BASED;
			addrtmp = (pldiidwbase - pldifdwbase) + pldiicount;
			if (checkwidth(addrtmp,9)!=TRUE) {
			    yyerror("error : PseudoLDII displacement out of range");
			    addrtmp = 0;
			}
			instword =
			    (inst.optype << 29) |
			    (inst.subtype1 << 26) |
			    (N_PSH << 8) |
			    ((addrtmp & 0x1ff) << 16) |
			    $4 ;
			pldiicount++;
		    }
		}
		incpc = 1;
		}
	| loadimm FLOATVAL COMMA sreg
		{
		int	addrtmp;

		if (pass == 1) {	/* pass 1 */
		    if (pldifcount>=MAXPLDIF) {
			yyerror("error : Pseudo LDIF buffer overflow");
			$2 = 0;
		    } else {
			pldifvalue[pldifcount].f = $2;
			pldifcount++;
		    }
		} else if (pass == 2) {	/* pass 2 */
		    /* PseudoLDI float = LOAD BASED(BASE=PSH)+DW */
		    inst.optype = LOAD;
		    inst.subtype1 = AD_BASED;
		    addrtmp = pldifcount;
		    if (checkwidth(addrtmp,9)!=TRUE) {
			yyerror("error : PseudoLDIF displacement out of range");
			addrtmp = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			(N_PSH << 8) |
			((addrtmp & 0x1ff) << 16) |
			$4 ;
		    pldifcount++;
		}
		incpc = 1;
		}
	| ls_ir intexp COMMA sreg
		{
		if (pass == 2) {
		    if (checkwidth($2,17)!=TRUE) {
			yyerror("error : LDI/LD/ST immediate value/address out of range");
			$2 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			(($2 & 0x1ffff) << 8) |
			$4 ;
		}
		incpc = 1;
		}
	| rel_ls_ir intexp COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			((reladdr($2,17) & 0x1ffff) << 8) |
			$4 ;
		}
		incpc = 1;
		}
	| ls_rir sreg COMMA intexp COMMA sreg
		{
		if (pass == 2) {
		    if (checkwidth($4,9)!=TRUE) {
			yyerror("error : LDB/STB displacement out of range");
			$4 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			($2 << 8) |
			(($4 & 0x1ff) << 16) |
			$6 ;
		}
		incpc = 1;
		}
	| ls_rir_shift sreg COMMA intexp COMMA sreg
		{
		if (pass == 2) {
		    if (($4 < 0)||($4 > 32)) {
			yyerror("error : LDBS/STBS invalid shift value");
			$4 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			($2 << 8) |
			(($4 & 0x1ff) << 16) |
			$6 ;
		}
		incpc = 1;
		}
	| ls_rrr sreg COMMA sreg COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			($2 << 16) |
			($4 << 8) |
			$6 ;
		}
		incpc = 1;
		}
	| ls_rr sreg COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			($2 << 16) |
			$4 ;
		}
		incpc = 1;
		}
	| ls_special
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 26) |
			((inst.src1 & 0x1ffff) << 8) |
			inst.dstn ;
		}
		incpc = 1;
		}
	;

loadimm
	: LDI
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_IMMED;
		}
	;

ls_ir
	: LD
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_DIRECT;
		}
	| ST
		{
		inst.optype = STORE;
		inst.subtype1 = AD_DIRECT;
		}
	;

rel_ls_ir
	: LDPCB
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_PCBASED;
		}
	| STPCB
		{
		inst.optype = STORE;
		inst.subtype1 = AD_PCBASED;
		}
	;
	

ls_rir
	: LDB
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_BASED;
		}
	| STB
		{
		inst.optype = STORE;
		inst.subtype1 = AD_BASED;
		}
	;

ls_rir_shift
	: LDBS
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_BASESFT;
		}
	| STBS
		{
		inst.optype = STORE;
		inst.subtype1 = AD_BASESFT;
		}
	;

ls_rrr
	: LDBI
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_BASEINDX;
		}
	| STBI
		{
		inst.optype = STORE;
		inst.subtype1 = AD_BASEINDX;
		}
	;

ls_rr
	: LDPCBI
		{
		inst.optype = LOAD;
		inst.subtype1 = AD_PCBINDX;
		}
	| STPCBI
		{
		inst.optype = STORE;
		inst.subtype1 = AD_PCBINDX;
		}
	;

ls_special
	: NOP
		{ /* NOP = LDI 0,ZR */
		inst.optype = LOAD;
		inst.subtype1 = AD_IMMED;
		inst.src1 = 0;
		inst.dstn = ZR;
		}
	| HLT
		{ /* HLT = LDI HALT,FR */
		inst.optype = LOAD;
		inst.subtype1 = AD_IMMED;
		inst.src1 = HALT;
		inst.dstn = FR;
		}
	;

RRRinst
	: RRRinsts sreg COMMA sreg COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			($2 << 16) |
			($4 << 8) |
			$6 ;
		}
		incpc = 1;
		}
	;

RRRinsts
	: intRRR
	| logicRRR
	| floatRRR
	;

intRRR
	: ADD
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_ADD;
		}
	| SUB
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_SUB;
		}
	| MUL
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_MUL;
		}
	| DIV
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_DIV;
		}
	| MOD
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_MOD;
		}
	| RSUB
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_RSUB;
		}
	| UMUL
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_UMUL;
		}
	| UDIV
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_UDIV;
		}
	;

logicRRR
	: AND
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_AND;
		}
	| OR
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_OR;
		}
	| XOR
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_XOR;
		}
	| SRL
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_SRL;
		}
	| SLL
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_SLL;
		}
	| SRA
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_SRA;
		}
	| SLA
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_SLA;
		}
	;

floatRRR
	: FADD
		{
		inst.optype = FLOATCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = FOP_ADD;
		}
	| FSUB
		{
		inst.optype = FLOATCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = FOP_SUB;
		}
	| FMUL
		{
		inst.optype = FLOATCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = FOP_MLT;
		}
	| FDIV
		{
		inst.optype = FLOATCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = FOP_DIV;
		}
	;

IRRinst
	: IRRinsts intexp COMMA sreg COMMA sreg
		{
		switch (inst.subtype2) {
		    case OP_UMUL:
		    case OP_UDIV:
			if (($2 < 0)||($2 > 255)) {
			    yyerror("unsigned immediate value out of range.");
			    $2 = 0;
			}
			break;
		    default:
			if (($2 < -128)||($2 > 127)) {
			    yyerror("immediate value out of range.");
			    $2 = 0;
			}
			break;
		}
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			(($2 & 0xff) << 16) |
			($4 << 8) |
			$6 ;
		}
		incpc = 1;
		}
	;

IRRinsts
	: intIRR
	| logicIRR
	;

intIRR
	: ADDI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_ADD;
		}
	| SUBI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SUB;
		}
	| MULI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_MUL;
		}
	| DIVI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_DIV;
		}
	| MODI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_MOD;
		}
	| RSUBI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_RSUB;
		}
	| UMULI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_UMUL;
		}
	| UDIVI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_UDIV;
		}
	;

logicIRR
	: ANDI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_AND;
		}
	| ORI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_OR;
		}
	| XORI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_XOR;
		}
	| SRLI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SRL;
		}
	| SLLI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SLL;
		}
	| SRAI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SRA;
		}
	| SLAI
		{
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SLA;
		}
	;

RRinst
	: RRinsts sreg COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			($2 << 16) |
			$4 ;
		}
		incpc = 1;
		}
	;

RRinsts
	: NOT
		{
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_NOT;
		}
	| INT
		{
		inst.optype = FLOATCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = FOP_INT;
		}
	| FLT
		{
		inst.optype = FLOATCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = FOP_FLT;
		}
	;

synthetic
	: MOV sreg COMMA sreg
		{ /* MOV SRs,SRd = OR SRs,ZR,SRd */
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_OR;
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			($2 << 16) |
			$4 ;
		}
		incpc = 1;
		}
	| CMP sreg COMMA sreg
		{ /* CMP SRs1,SRs2 = SUB SRs1,SRs2,ZR */
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_SUB;
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			($2 << 16) |
			($4 << 8) ;
		}
		incpc = 1;
		}
	| CMPI intexp COMMA sreg
		{ /* CMPI Immed,SRs2 = SUBI Immed,SRs2,ZR */
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SUB;
		if (($2 < -128)||($2 > 127)) {
			yyerror("immediate value out of range.");
			$2 = 0;
		}
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			(($2 & 0xff) << 16) |
			($4 << 8) ;
		}
		incpc = 1;
		}
	| RCMP sreg COMMA sreg
		{ /* RCMP SRs1,SRs2 = RSUB SRs1,SRs2,ZR */
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_RSUB;
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			($2 << 16) |
			($4 << 8) ;
		}
		incpc = 1;
		}
	| RCMPI intexp COMMA sreg
		{ /* RCMPI Immed,SRs2 = RSUBI Immed,SRs2,ZR */
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_RSUB;
		if (($2 < -128)||($2 > 127)) {
			yyerror("immediate value out of range.");
			$2 = 0;
		}
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			(($2 & 0xff) << 16) |
			($4 << 8) ;
		}
		incpc = 1;
		}
	| INC	sreg
		{ /* INC SRd = ADDI 1,SRd,SRd */
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_ADD;
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			(1 << 16) |
			($2 << 8) |
			$2 ;
		}
		incpc = 1;
		}
	| DEC	sreg
		{ /* DEC SRd = SUBI 1,SRd,SRd */
		inst.optype = INTCALC;
		inst.subtype1 = IMMOPR;
		inst.subtype2 = OP_SUB;
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			(1 << 16) |
			($2 << 8) |
			$2 ;
		}
		incpc = 1;
		}
	| CLR	sreg
		{ /* CLR SRd = OR SR0,SR0,SRd */
		inst.optype = INTCALC;
		inst.subtype1 = REGOPR;
		inst.subtype2 = OP_OR;
		if (pass == 2) {
		    instword = 
			(inst.optype << 29) |
			(inst.subtype1 << 28) |
			(inst.subtype2 << 24) |
			$2 ;
		}
		incpc = 1;
		}
	;

jump
	: jumpi intexp
		{
		if (pass == 2) {
		    if (($2 % NIP) != 0) {
			yyerror("Jump address alignment error");
		    }
		    $2 = reladdr($2,24+PADSFT)>>PADSFT;
		    if (checkwidth($2,24)!=TRUE) {
			yyerror("error : JMP/JMPL address out of range");
			$2 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 24) |
			($2 & 0xffffff);
		}
		incpc = 1;
		}
	| jumpri sreg COMMA intexp
		{
		if (pass == 2) {
		    if (($4 & PADMASK) != $4) {
			yyerror("LSB(s) will be ignored.");
		    }
		    $4 = $4 >> PADSFT;
		    if (checkwidth($4,16)!=TRUE) {
			yyerror("error : JB/JBL displacement out of range");
			$4 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 24) |
			($2 << 16) |
			($4 & 0xffff) ;
		}
		incpc = 1;
		}
	| jumpci cond COMMA intexp
		{
		if (pass == 2) {
		    if (($4 % NIP) != 0) {
			yyerror("Jump address alignment error");
		    }
		    $4 = reladdr($4,24+PADSFT)>>PADSFT;
		    if (checkwidth($4,24)!=TRUE) {
			yyerror("error : JC/JCL address out of range");
			$4 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 24) |
			($4 & 0xffffff) ;
		}
		incpc = 1;
		}
	| jumpcri cond COMMA sreg COMMA intexp
		{
		if (pass == 2) {
		    if (($6 & PADMASK) != $6) {
			yyerror("LSB(s) will be ignored.");
		    }
		    $6 = $6 >> PADSFT;
		    if (checkwidth($6,16)!=TRUE) {
			yyerror("error : JBC/JBCL displacement out of range");
			$6 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 24) |
			($4 << 16) |
			($6 & 0xffff) ;
		}
		incpc = 1;
		}
	;

jumpi
	: JMP
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_REL;
		inst.subtype2 = JC_NONCND;
		}
	| JMPL
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_REL_LK;
		inst.subtype2 = JC_NONCND;
		}
	;

jumpri
	: JB
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_BASED;
		inst.subtype2 = JC_NONCND;
		}
	| JBL
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_BASED_LK;
		inst.subtype2 = JC_NONCND;
		}
	;

jumpci
	: JC
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_REL;
		}
	| JCL
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_REL_LK;
		}
	;

jumpcri
	: JBC
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_BASED;
		}
	| JBCL
		{
		inst.optype = JUMP;
		inst.subtype1 = JAD_BASED_LK;
		}
	;

cond
	: P
		{
		inst.subtype2 = JC_PLUS;
		}
	| M
		{
		inst.subtype2 = JC_MINUS;
		}
	| Z
		{
		inst.subtype2 = JC_ZERO;
		}
	| NP
		{
		inst.subtype2 = JC_NOTPLUS;
		}
	| NM
		{
		inst.subtype2 = JC_NOTMINUS;
		}
	| NZ
		{
		inst.subtype2 = JC_NOTZERO;
		}
	;

compjump
	: cjrr CJops COMMA sreg COMMA sreg COMMA intexp
		{
		if (pass == 2) {
		    if (($8 % NIP) != 0) {
			yyerror("Jump address alignment error");
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 26) |
			(inst.subtype3 << 24) |
			($4 << 16) |
			($6 << 8) |
			((reladdr($8,8+PADSFT)>>PADSFT) & 0xff) ;
		}
		incpc = 1;
		}
	| cjir CJops COMMA intexp COMMA sreg COMMA intexp
		{
		if (pass == 2) {
		    if (($8 % NIP) != 0) {
			yyerror("Jump address alignment error");
		    }
		    if (checkwidth($4,8)!=TRUE) {
			yyerror("error : Comp&jump immediate value out of range");
			$4 = 0;
		    }
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 26) |
			(inst.subtype3 << 24) |
			(($4 & 0xff) << 16) |
			($6 << 8) |
			((reladdr($8,8+PADSFT)>>PADSFT) & 0xff) ;
		}
		incpc = 1;
		}
	;

cjrr
	: CJP
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_PLUS;
		inst.subtype2 = REGOPR;
		}
	| CJM
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_MINUS;
		inst.subtype2 = REGOPR;
		}
	| CJZ
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_ZERO;
		inst.subtype2 = REGOPR;
		}
	| CJNZ
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_NONZERO;
		inst.subtype2 = REGOPR;
		}
	;

cjir
	: CJPI
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_PLUS;
		inst.subtype2 = IMMOPR;
		}
	| CJMI
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_MINUS;
		inst.subtype2 = IMMOPR;
		}
	| CJZI
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_ZERO;
		inst.subtype2 = IMMOPR;
		}
	| CJNZI
		{
		inst.optype = COMPJUMP;
		inst.subtype1 = CJC_NONZERO;
		inst.subtype2 = IMMOPR;
		}
	;

CJops
	: SUB
		{
		inst.subtype3 = CJO_SUB;
		}
	| RSUB
		{
		inst.subtype3 = CJO_RSUB;
		}
	| AND
		{
		inst.subtype3 = CJO_AND;
		}
	| OR
		{
		inst.subtype3 = CJO_OR;
		}
	;

vcalc
	: v_rrr VR COMMA VR COMMA VR usedvmr
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 22) |
			(inst.vmr << 19) |
			(inst.subtype3 << 18) |
			($2 << 12) |
			($4 << 6) |
			$6 ;
		}
		incpc = 1;
		}
	| VNOT VR COMMA VR usedvmr
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LNOT;
		inst.subtype3 = 0; /* chain flag OFF */
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 22) |
			(inst.vmr << 19) |
			(inst.subtype3 << 18) |
			($2 << 12) |
			$4 ;
		}
		incpc = 1;
		}
	;

v_rrr
	: VFADD
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FADD;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VFSUB
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FSUB;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VFMUL
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FMULT;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VFDIV
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FDIV;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VADD
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_IADD;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VSUB
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_ISUB;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VMUL
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_IMULT;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VDIV
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_IDIV;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VAND
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LAND;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VOR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LOR;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	| VXOR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LXOR;
		inst.subtype3 = 0; /* chain flag OFF */
		}
	;

usedvmr
	: /* empty */
		{
		inst.vmr = 0;
		}
	| COMMA VMR
		{
		inst.vmr = $2 & 0x7;
		}
	;

cvcalc
	: cv_rrr vrcn COMMA vrcn COMMA vrcn usedvmr
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 22) |
			(inst.vmr << 19) |
			(inst.subtype3 << 18) |
			makecvinst($2,$4,$6) ;
		}
		incpc = 1;
		}
	| CVNOT vrcn COMMA vrcn usedvmr
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LNOT;
		inst.subtype3 = 1; /* chain flag ON */
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 22) |
			(inst.vmr << 19) |
			(inst.subtype3 << 18) |
			makecvinst($2,0,$4) ;
		}
		incpc = 1;
		}
	;

cv_rrr
	: CVFADD
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FADD;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVFSUB
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FSUB;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVFMUL
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FMULT;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVFDIV
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_FDIV;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVADD
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_IADD;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVSUB
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_ISUB;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVMUL
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_IMULT;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVDIV
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_IDIV;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVAND
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LAND;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVOR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LOR;
		inst.subtype3 = 1; /* chain flag ON */
		}
	| CVXOR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_LXOR;
		inst.subtype3 = 1; /* chain flag ON */
		}
	;

vrcn
	: VR
		{
		$$ = $1;
		}
	| CN
		{
		$$ = $1 + CNOFFSET;
		}
	;

vsmixed
	: VBCAST sr128 COMMA VR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_BCAST;
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 22) |
			(inst.vmr << 19) |
			($2 << 12) |
			$4 ;
		}
		incpc = 1;
		}
	| vshift sr128 COMMA VR COMMA VR
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 22) |
			(inst.vmr << 19) |
			($2 << 12) |
			($4 << 6) |
			$6 ;
		}
		incpc = 1;
		}
	;

sr128
	: sreg
		{
		if (($1 < 0)||($1 > 127)) { /* SR value range check */
			yyerror("error : VBCAST/VSRL/VSLL SR out of range");
			$$ = 0;
		} else { /* check OK! */
			$$ = $1;
		}
		}
	;

vshift
	: VSRL
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_VSRL;
		}
	| VSLL
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_CALC;
		inst.subtype2 = VOP_VSLL;
		}
	;

vls
	: vlsir VR COMMA sreg COMMA sreg usedvmr
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 25) |
			(inst.vmr << 22) |
			($2 << 16) |
			($4 << 8) |
			$6 ;
		}
		incpc = 1;
		}
	| vlsii VR COMMA sreg COMMA intexp usedvmr
		{
		if (($6 <= 0)||($6 > 255)) {
		    yyerror("VLII/VSII increment out of range.");
		    $6 = 0;
		}
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 25) |
			(inst.vmr << 22) |
			($2 << 16) |
			($4 << 8) |
			$6 ;
		}
		incpc = 1;
		}
	;

vlsir
	: VLIR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_LS;
		inst.subtype2 = VOP_LDIR;
		}
	| VSIR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_LS;
		inst.subtype2 = VOP_STIR;
		}
	;

vlsii
	: VLII
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_LS;
		inst.subtype2 = VOP_LDII;
		}
	| VSII
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_LS;
		inst.subtype2 = VOP_STII;
		}
	;

lvls
	: lvldst VR COMMA VR COMMA usedvmr
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			(inst.vmr << 13) |
			(inst.subtype3 << 12) |
			($2 << 6) |
			$4 ;
		}
		incpc = 1;
		}
	| lvbldst VR COMMA VR COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			($6 << 13) |
			(inst.subtype3 << 12) |
			($2 << 6) |
			$4 ;
		}
		incpc = 1;
		}
	;

lvldst
	: LVLD
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_LVLS;
		inst.subtype3 = 0; /* LOAD = 0 */
		}
	| LVST
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_LVLS;
		inst.subtype3 = 1; /* STORE = 1 */
		}
	;

lvbldst
	: LVBLD
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_LVBLS;
		inst.subtype3 = 0; /* LOAD = 0 */
		}
	| LVBST
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_LVBLS;
		inst.subtype3 = 1; /* STORE = 1 */
		}
	;

vmrset
	: vmrsetif vmrsetcond COMMA VR COMMA VMR
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			(inst.subtype3 << 9) |
			($4 << 3) |
			$6 ;
		}
		incpc = 1;
		}
	;

vmrsetif
	: VMRSETF
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_VMRSET;
		inst.subtype3 = 0; /* compare as float = 0 */
		}
	| VMRSETI
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_VMRSET;
		inst.subtype3 = 8; /* compare as int = 8 */
		}
	;

vmrsetcond
	: P
		{
		inst.subtype3 |= MC_F_PLUS;
		}
	| M
		{
		inst.subtype3 |= MC_F_MINUS;
		}
	| Z
		{
		inst.subtype3 |= MC_F_ZERO;
		}
	| NP
		{
		inst.subtype3 |= MC_F_NOTPLUS;
		}
	| NM
		{
		inst.subtype3 |= MC_F_NOTMINUS;
		}
	| NZ
		{
		inst.subtype3 |= MC_F_NOTZERO;
		}
	;

vmrop
	: vmr_rrr VMR COMMA VMR COMMA VMR
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			(inst.subtype3 << 9) |
			($2 << 6) |
			($4 << 3) |
			$6 ;
		}
		incpc = 1;
		}
	| VMRNOT VMR COMMA VMR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_VMROP;
		inst.subtype3 = MOP_NOT;
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			(inst.subtype3 << 9) |
			($2 << 6) |
			$4 ;
		}
		incpc = 1;
		}
	;

vmr_rrr
	: VMRAND
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_VMROP;
		inst.subtype3 = MOP_AND;
		}
	| VMROR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_VMROP;
		inst.subtype3 = MOP_OR;
		}
	| VMRXOR
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_VMROP;
		inst.subtype3 = MOP_XOR;
		}
	;

vconv
	: vconvop VR COMMA VR usedvmr
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			(inst.vmr << 13) |
			(inst.subtype3 << 12) |
			($2 << 6) |
			$4 ;
		}
		incpc = 1;
		}
	;

vconvop
	: VINT
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_CONV;
		inst.subtype3 = VOP_CONVF2I;
		}
	| VFLT
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_CONV;
		inst.subtype3 = VOP_CONVI2F;
		}
	;

vsmove
	: vsmoveop VR COMMA sreg COMMA sreg
		{
		if (pass == 2) {
		    instword =
			(inst.optype << 29) |
			(inst.subtype1 << 27) |
			(inst.subtype2 << 23) |
			($2 << 16) |
			($4 << 8) |
			$6 ;
		}
		incpc = 1;
		}
	;

vsmoveop
	: MOVVS
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_MOVEV2S;
		}
	| MOVSV
		{
		inst.optype = VECTOR;
		inst.subtype1 = V_MISC;
		inst.subtype2 = VOP_MOVES2V;
		}
	;

intexp
	: LABEL
		{
		$$ = getlabelvalue($1);
		}
	| INTVAL
		{
		$$ = $1;
		}
	| LP intexp RP
		{
		$$ = $2;
		}
	| intexp ADDOP intexp
		{
		$$ = $1 + $3;
		}
	| intexp SUBOP intexp
		{
		$$ = $1 - $3;
		}
	| intexp MULOP intexp
		{
		$$ = $1 * $3;
		}
	| intexp DIVOP intexp
		{
		if ($3 != 0) 
		    $$ = $1 / $3;
		else {
		    if (uninitlabelerr == TRUE) {
			yyerror("divide by zero in integer expression");
		    }
		    $$ = 0;
		}
		}
	| intexp MODOP intexp
		{
		if ($3 != 0) 
			$$ = $1 % $3;
		else {
		    if (uninitlabelerr == TRUE) {
			yyerror("modulo by zero in integer expression");
		    }
		    $$ = 0;
		}
		}
	| intexp RSFTOP intexp
		{
		$$ = $1 >> $3;
		}
	| intexp LSFTOP intexp
		{
		$$ = $1 << $3;
		}
	| intexp ANDOP intexp
		{
		$$ = $1 & $3;
		}
	| intexp OROP intexp
		{
		$$ = $1 | $3;
		}
	| intexp XOROP intexp
		{
		$$ = $1 ^ $3;
		}
	| ADDOP intexp %prec UPLUS
		{
		$$ = $2;
		}
	| SUBOP intexp %prec UMINUS
		{
		$$ = -$2;
		}
	| NOTOP intexp
		{
		$$ = ~$2;
		}
	;
%%
/*
   search label , and return its pointer.
   if not found , add label list.
   (labelvalue is not initialized)
*/
labelnode *addlabel(name)
char *name;
{
	labelnode	*lnp,*lastlnp,*newlnp;
#ifdef DEBUG
	fprintf(stderr,"addlabel %s\n",name);
#endif

	if (root == NULL) { /* no label added */
		root = labelnodealloc();
		root->labelname = name;
		root->labelvalue = 0;
		root->initialized = FALSE;
		root->next = NULL;
		return root;
	} else { /* already some labels added */ 
		lnp = root;
		while (lnp != NULL) {
			if (strcmp(lnp->labelname,name) == 0) {
				return (lnp);
			}
			lastlnp = lnp;
			lnp = lnp->next;
		}
		if (lnp == NULL) { /* not found */
			newlnp = labelnodealloc();
			lastlnp->next = newlnp;
			newlnp->labelname = name;
			newlnp->labelvalue = 0;
			newlnp->initialized = FALSE;
			newlnp->next = NULL;
			return newlnp;
		}
	}
}

/*
   search label , set label value and return its pointer.
   if not found , add label tree and set label value.
*/
labelnode *addlabelandinit(name,value)
char	*name;
int	value;
{
	labelnode *foundlabel;
	char	errorstr[256];

#ifdef DEBUG
	fprintf(stderr,"addlabelandinit %s,%d\n",name,value);
#endif
	foundlabel = addlabel(name);

	foundlabel->labelvalue = value;
#ifdef DEBUG
	fprintf(stderr,"addlabelandinit label=%s value=(%d)\n",name,foundlabel->labelvalue);
#endif

	if (foundlabel->initialized == TRUE) {
		sprintf(errorstr,"warning: label redefined. (%s)\n",
			foundlabel->labelname);
		yyerror(errorstr);
	}
	foundlabel->initialized = TRUE;
	
	return foundlabel;
}

/*
   search label , and get it's value.
   if not found or not initialized, returns 0.
*/
int getlabelvalue(name)
char	*name;
{
	labelnode *foundlabel;
	int	returnvalue;
	char	errorstr[256];

#ifdef DEBUG
	fprintf(stderr,"getlabelvalue %s\n",name);
#endif
	foundlabel = addlabel(name);
#ifdef DEBUG
	fprintf(stderr,"getlabelvalue label=%s value=(%d)\n",name,foundlabel->labelvalue);
#endif

	if (foundlabel->initialized == TRUE) {
	    returnvalue = foundlabel->labelvalue;
	} else { /* labelvalue is not initialized */
	    if (uninitlabelerr == TRUE) {
		/* uninitlabelused = TRUE; */
		sprintf(errorstr,"warning: uninitialized label used. (%s)\n",
			foundlabel->labelname);
		yyerror(errorstr);
	    }
	    returnvalue = 0;
	}
	
	return returnvalue;
}

/*
   allocate memory for labelnode
*/
labelnode *labelnodealloc()
{
	return (labelnode *) malloc(sizeof(labelnode));
}

/* printout yacc error message */
void yyerror(s)
char *s;
{
	extern int yynerrs,yylineno;

	++yynerrs;
	fprintf(stderr,"%d: %s\n",yylineno,s);
	listoutcmnt(s);
	errorflag = TRUE;
}

/* calculation relative address */
int reladdr(target,bitsize)
int	target,bitsize;
{
	int	disp;	/* displacement */
	extern	int	currentpc;	/* current PC */

	disp = target - (currentpc & PADMASK) - NIP;

	if (checkwidth(disp,bitsize) == FALSE) {
		yyerror("relative address out of range!");
		disp = 0;
	}

	return disp;
}

/* check bit width (OK = return TRUE , else return FALSE) */
int	checkwidth(value,width)
int	value,width;
{
	int check;

	if (value < 0) {
		check = (value & bitmask[width])|(~bitmask[width]);
	} else {
		check = value & bitmask[width];
	}

	if (value != check) return FALSE;
	else return TRUE;
}

/* make chained vector instruction operands field */
int	makecvinst(src1,src2,dstn)
int	src1;	/* source 1 */
int	src2;	/* source 2 */
int	dstn;	/* destination */
{
	int	operand;
	int	ctype;

	if (src1 >= CNOFFSET) { /* source 1 is CN */
	    if (src2 >= CNOFFSET) { /* src1,src2 = CN */
		if (dstn >= CNOFFSET) {
		    ctype = CHAIN_CCC; /* src1=CN,src2=CN,dstn=CN */
		} else {
		    ctype = CHAIN_CCR; /* src1=CN,src2=CN,dstn=VR */
		}
	    } else { /* src1 = CN, src2 = VR */
		if (dstn >= CNOFFSET) {
		    ctype = CHAIN_CRC; /* src1=CN,src2=VR,dstn=CN */
		} else {
		    ctype = CHAIN_CRR; /* src1=CN,src2=VR,dstn=VR */
		}
	    }
	} else { /* source 1 is VR */
	    if (src2 >= CNOFFSET) { /* src1 = VR , src2 = CN */
		if (dstn >= CNOFFSET) {
		    ctype = CHAIN_RCC; /* src1=VR,src2=CN,dstn=CN */
		} else {
		    ctype = CHAIN_RCR; /* src1=VR,src2=CN,dstn=VR */
		}
	    } else { /* src1 = VR, src2 = VR */
		if (dstn >= CNOFFSET) {
		    ctype = CHAIN_RRC; /* src1=VR,src2=VR,dstn=CN */
		} else {
		    ctype = CHAIN_OFF; /* src1=VR,src2=VR,dstn=VR */
		}
	    }
	}

	switch (ctype) {
	    case CHAIN_OFF:
		yyerror("invalid chaining instruction");
		operand = 0x38000; /* CHAIN_CCC,0,0,0 */
		break;
	    case CHAIN_RRC:
		dstn -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 9 ) |
			  (src2 << 3 ) |
			  dstn ;
		break;
	    case CHAIN_RCR:
		src2 -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 9 ) |
			  (src2 << 6 ) |
			  dstn ;
		break;
	    case CHAIN_RCC:
		src2 -= CNOFFSET;
		dstn -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 9 ) |
			  (src2 << 6 ) |
			  dstn ;
		break;
	    case CHAIN_CRR:
		src1 -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 12 ) |
			  (src2 << 6 ) |
			  dstn ;
		break;
	    case CHAIN_CRC:
		src1 -= CNOFFSET;
		dstn -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 12 ) |
			  (src2 << 6 ) |
			  dstn ;
		break;
	    case CHAIN_CCR:
		src1 -= CNOFFSET;
		src2 -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 12 ) |
			  (src2 << 6 ) |
			  dstn ;
		break;
	    case CHAIN_CCC:
		src1 -= CNOFFSET;
		src2 -= CNOFFSET;
		dstn -= CNOFFSET;
		operand = (ctype << 15) |
			  (src1 << 12 ) |
			  (src2 << 6 ) |
			  dstn ;
		break;
	}

	return operand;
}
