/* pipe-sim(+vector) ver.4.0 [92/05/30-06/15] */
/* define opcode constants */

/* opcode defines */

/* instruction type */

#define NO_OPR  -1		/* do nothing */
#define LOAD	0		/* load from memory */
#define STORE	1		/* store to memory */

#define INTCALC	2		/* integer calculation */
#define FLOATCALC	3	/* float calculation */

#define JUMP	4		/* jump */
#define COMPJUMP	5	/* compare and jump */

#define VECTOR	6		/* vector instruction */

/* LOAD/STORE instruction addressing mode */

#define AD_IMMED	0	/* immediate */
#define AD_DIRECT	1	/* direct */
#define AD_BASED	2	/* based / indexed */
#define AD_BASEINDX	3	/* base-indexed */
#define AD_BASESFT	4	/* base-shifted */
#define AD_PCBASED	5	/* PC-based */
#define AD_PCBINDX	6	/* PC-base-indexed */

/* ALU operand immediate flag */

#define REGOPR	0	/* register operand */
#define IMMOPR	1	/* immediate operand */

/* ALU operation */
#define OP_ADD	0
#define OP_SUB	1
#define OP_MUL	2
#define OP_DIV	3
#define OP_MOD	4	/* modulo operation */
#define OP_RSUB	5	/* reverse operand subtract */
#define OP_UMUL	6	/* unsigned mult */
#define OP_UDIV	7	/* unsigned div */
#define OP_AND	8
#define OP_OR	9
#define OP_XOR	10
#define OP_NOT	11
#define OP_SRL	12	/* shift right logical */
#define OP_SLL	13	/* shift left logical */
#define OP_SRA	14	/* shift right arithmetic */
#define OP_SLA	15	/* shift left arithmetic */

#define OP_NOP -1	/* no operation (for int/logic pipes) */

/* float operation */
#define FOP_ADD	0
#define FOP_SUB	1
#define FOP_MLT	2
#define FOP_DIV	3

#define FOP_INT	8	/* float to int conversion */
#define FOP_FLT	9	/* int to float conversion */

#define FOP_NOP -1	/* no operation (for floating pipes) */

/* jump addressing mode */
#define JAD_REL		0	/* relative */
#define JAD_BASED	1	/* register-based */
#define JAD_REL_LK	2	/* relative & link */
#define JAD_BASED_LK	3	/* register-based & link */

/* jump conditions */
#define JC_NONCND	0	/* non conditional jump */
#define JC_PLUS		1
#define JC_MINUS	2
#define JC_ZERO		3
#define JC_NOTPLUS	4
#define JC_NOTMINUS	5
#define JC_NOTZERO	6

/* compare&jump conditions */
#define CJC_PLUS	0
#define CJC_MINUS	1
#define CJC_ZERO	2
#define CJC_NONZERO	3

/* compare&jump operations */
#define CJO_SUB		0
#define CJO_RSUB	1
#define CJO_AND		2
#define CJO_OR		3
#define CJO_IMSUB	4	/* immed subtract */
#define CJO_IMRSUB	5	/* immed reverse-subtract */
#define CJO_IMAND	6	/* immed and */
#define CJO_IMOR	7	/* immed or */

/* vector operation */
#define V_CALC	0	/* vector calculation (float,int,logic,vsmix) */
#define V_LS	1	/* vector load/store */
#define V_MISC	2	/* mask reg, list vector L/S, f-i conv, etc. */

/* vector calc operations */
/* floating operation */
#define VOP_FADD	0
#define VOP_FSUB	1
#define VOP_FMULT	2
#define VOP_FDIV	3
/* integer operation */
#define VOP_IADD	4
#define VOP_ISUB	5
#define VOP_IMULT	6
#define VOP_IDIV	7
/* integer loogical operation */
#define VOP_LAND	8
#define VOP_LOR		9
#define VOP_LXOR	10
#define VOP_LNOT	11
/* vector-scalar mixed operation */
#define VOP_BCAST	12	/* scalar value broadcast */
   /*  not defined   ->	13 */
#define VOP_VSRL	14
#define VOP_VSLL	15
/* chaining control */
#define	CHAIN_OFF	0	/* Chaining not used */
#define	CHAIN_RRC	1	/* src1=Reg  ,src2=Reg  ,dstn=Chain */
#define	CHAIN_RCR	2	/* src1=Reg  ,src2=Chain,dstn=Reg   */
#define	CHAIN_RCC	3	/* src1=Reg  ,src2=Chain,dstn=Chain */
#define	CHAIN_CRR	4	/* src1=Chain,src2=Reg  ,dstn=Reg   */
#define	CHAIN_CRC	5	/* src1=Chain,src2=Reg  ,dstn=Chain */
#define	CHAIN_CCR	6	/* src1=Chain,src2=Chain,dstn=Reg   */
#define	CHAIN_CCC	7	/* src1=Chain,src2=Chain,dstn=Chain */

/* vector load/store */
/* addressing mode is based-autoincrement only. */
/* increment value is in scalar reg or immediate constant. */
#define VOP_LDIR	0	/* vector load, increment is s-reg. */
#define VOP_LDII	1	/* vector load, increment is immed. */
#define VOP_STIR	2	/* vector store, increment is s-reg. */
#define VOP_STII	3	/* vector store, increment is immed. */

/* vector mask reg.(VMR) instructions */
#define VOP_VMROP	0	/* VMR operation */
#define VOP_VMRSET	1	/* set VMR */
/* VMR operations */
#define MOP_AND		0
#define MOP_OR		1
#define MOP_XOR		2
#define MOP_NOT		3
/* VMR set conditions(compare as float) */
#define MC_F_ZERO	0
#define MC_F_NOTZERO	1
#define MC_F_PLUS	2
#define MC_F_NOTPLUS	3
#define MC_F_MINUS	4
#define MC_F_NOTMINUS	5
/* VMR set conditions(compare as integer) */
#define MC_I_ZERO	8
#define MC_I_NOTZERO	9
#define MC_I_PLUS	10
#define MC_I_NOTPLUS	11
#define MC_I_MINUS	12
#define MC_I_NOTMINUS	13

/* list vector load/store */
#define VOP_LVLS	2	/* list vector load/store */
#define VOP_LVBLS	3	/* list vector+base reg. load/store */

#define VOP_LVLD	4	/* list vector load */
#define VOP_LVBLD	5	/* list vector+base reg. load */
#define VOP_LVST	6	/* list vector store */
#define VOP_LVBST	7	/* list vector+base reg. store */

/* vectore reg. float/int conversion */
#define VOP_CONV	4

#define VOP_CONVI2F	0	/* int to float conv. */
#define VOP_CONVF2I	1	/* float to int conv. */

/* vector reg <-> scalar reg data move */
#define VOP_MOVES2V	6	/* move scalar to vector */
#define VOP_MOVEV2S	7	/* move vector to scalar */

/* vector int/logic pipe(VILP) opcodes */
#define VILPOP_ADD	0	/* integer add */
#define VILPOP_SUB	1	/* integer sub */
#define VILPOP_MUL	2	/* integer mult */
#define VILPOP_DIV	3	/* integer div */
#define VILPOP_AND	4	/* logical and */
#define VILPOP_OR	5	/* logical or  */
#define VILPOP_XOR	6	/* logical xor */
#define VILPOP_NOT	7	/* logical not */
#define VILPOP_SLL	8	/* shift left logical */
#define VILPOP_SRL	9	/* shift right logical */
#define VILPOP_BC	10	/* scalar value broadcast */
#define VILPOP_I2F	11	/* convert int to float */
#define VILPOP_F2I	12	/* convert float to int */
#define VILPOP_ICZ	13	/* integer compare to zero */
#define VILPOP_FCZ	14	/* float compare to zero */

#define VILPOP_NOP	-1	/* no operation */
